"""
Classes to serialize and deserialize custom types used by openedx events.
"""
from abc import ABC, abstractmethod
from datetime import datetime

from opaque_keys.edx.keys import CourseKey

from openedx_events.bridge.avro_types import AVRO_TYPE_FOR


class AvroAttrsBridgeExtention(ABC):
    """
    Used by openedx_events.avro_attrs_bridge.AvroAttrsBridge class to serialize/deserialize custom types.
    """

    cls: type

    def type(self):
        """Get type of class this extension serilizer and deserializes."""
        return type(self.cls)

    @abstractmethod
    def serialize(self, obj) -> str:
        """Abstart method to serialize obj into string."""
        ...

    @abstractmethod
    def deserialize(self, data: str) -> object:
        """Abstart method to deserialize string array into obj."""
        ...

    @abstractmethod
    def record_fields(self):
        """
        Abstract method to define avro schema for self.cls.

        This is usually just a AVRO_TYPE_FOR[str]
        """
        ...


class CourseKeyAvroAttrsBridgeExtension(AvroAttrsBridgeExtention):
    """
    AvroAttrsBrdgeExtension for CourseKey class.
    """

    cls = CourseKey

    def serialize(self, obj) -> str:
        """Serialize obj into string."""
        return str(obj)

    def deserialize(self, data: str):
        """Deserialize string array into obj."""
        return CourseKey.from_string(data)

    def record_fields(self):
        """Define avro schema for self.cls."""
        return AVRO_TYPE_FOR[str]


class DatetimeAvroAttrsBridgeExtension(AvroAttrsBridgeExtention):
    """
    AvroAttrsBrdgeExtension for CourseKey class.
    """

    cls = datetime

    def serialize(self, obj) -> str:
        """Serialize obj into string."""
        return obj.isoformat()

    def deserialize(self, data: str):
        """Deserialize string array into obj."""
        return datetime.fromisoformat(data)

    def record_fields(self):
        """Define avro schema for self.cls."""
        return AVRO_TYPE_FOR[str]



"""A mapping of python types to the avro type that we want to use make valid avro schema."""
AVRO_TYPE_FOR = {
    None: "null",
    bool: "boolean",
    int: "long",
    float: "double",
    bytes: "bytes",
    str: "string",
    dict: "record",
    list: "array",
}



"""
Code to convert attr classes to avro specification.
"""
import io
import json
import uuid
from datetime import datetime
from typing import Any, Dict

import attr
import fastavro

from openedx_events.bridge.avro_attrs_bridge_extensions import DatetimeAvroAttrsBridgeExtension
from openedx_events.bridge.avro_types import AVRO_TYPE_FOR


class AvroAttrsBridge:
    """
    Use to convert between Avro and Attrs data specifications.

    Intended usecase: To abstract serilalization and deserialization of openedx-events to send over pulsar or kafka
    """

    # default extensions, can be overwriteen by passing in extensions during obj initialization
    default_extensions = {
        DatetimeAvroAttrsBridgeExtension.cls: DatetimeAvroAttrsBridgeExtension()
    }
    # default config, should be overwritten by passing in config during obj initialization
    default_config = {
        "source": "/openedx/unknown/avro_attrs_bridge",
        "sourcehost": "unknown",
        "type": "org.openedx.test.test.test.v0",
    }

    def __init__(self, attrs_cls, extensions=None, config=None):
        """
        Init method for Avro Attrs Bridge.

        Arguments:
            attrs_cls: Attr Class Object (not instance)
            extensions: dict mapping Class Object to its AvroAttrsBridgeExtention subclass instance
            config: dict with followings keys
                - source:  This field will be used to indicate the logical source of an event, and will be of the form
                           /{namespace}/{service}/{web|worker}.
                           All services in standard distribution of Open edX should use openedx for the namespace.
                           Examples of services might be “discovery”, “lms”, “studio”, etc.
                           The value “web” will be used for events emitted by the web application,
                           and “worker” will be used for events emitted by asynchronous tasks such as celery workers.
                           For more info, see OEP-41: Asynchronous Server Event Message Format
                - sourcehost: should represent the physical source of message
                               -- i.e. host identifier of the server that emitted this event (example: edx.devstack.lms)
                              For more info, see OEP-41: Asynchronous Server Event Message Format
                - type: The name of event.
                        Should be formatted `{Reverse DNS}.{Architecture Subdomain}.{Subject}.{Action}.{Major Version}`.
                        For more info, see OEP-41: Asynchronous Server Event Message Format
        """
        self._attrs_cls = attrs_cls

        self.extensions = {}
        self.extensions.update(self.default_extensions)
        if isinstance(extensions, dict):
            self.extensions.update(extensions)

        self.config = {}
        self.config.update(self.default_config)
        if isinstance(config, dict):
            self.config.update(config)

        # Used by record_field_for_attrs_class function to keep track of which
        # records have already been defined in schema.
        # Reason: fastavro does no allow you to define record with same name twice
        self.schema_record_names = set()
        self.schema_dict = self._attrs_to_avro_schema(attrs_cls)
        # make sure the schema is parsable
        fastavro.parse_schema(self.schema_dict)

    def schema_str(self):
        """Json dumps schema dict into a string."""
        return json.dumps(self.schema_dict, sort_keys=True)

    def _attrs_to_avro_schema(self, attrs_cls):
        """
        Generate avro schema for attr_cls.

        Arguments:
            attrs_cls: Attr class object
        Returns:
            complex dict that defines avro schema for attrs_cls
        """
        base_schema = {
            "namespace": "io.cloudevents",
            "type": "record",
            "name": "CloudEvent",
            "version": "1.0",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/avro_attrs_bridge",
            "fields": [
                {"name": "id", "type": "string"},
                {"name": "type", "type": "string"},
                {"name": "specversion", "type": "string", "default": "1.0"},
                {"name": "time", "type": "string"},
                {"name": "source", "type": "string"},
                {"name": "sourcehost", "type": "string"},
                {"name": "minorversion", "type": "int"},
            ],
        }

        record_fields = self._record_fields_for_attrs_class(attrs_cls)
        base_schema["fields"].append(record_fields)
        return base_schema

    def _record_fields_for_attrs_class(
        self, attrs_class, field_name: str = "data"
    ) -> Dict[str, Any]:
        """
        Generate avro record for attrs_class.

        Will also recursively generate avro records for any sub attr classes and any custom types.

        Custom types are handled by AvroAttrsBridgeExtention subclass instance values defined in self.extensions dict
        """
        field: Dict[str, Any] = {}
        field["name"] = field_name
        field["type"] = dict(name=attrs_class.__name__, type="record", fields=[])

        for attribute in attrs_class.__attrs_attrs__:
            # Attribute is a simple type.
            if attribute.type in AVRO_TYPE_FOR:
                inner_field = {
                    "name": attribute.name,
                    "type": AVRO_TYPE_FOR[attribute.type],
                }

            # Attribute is another attrs class
            elif hasattr(attribute.type, "__attrs_attrs__"):
                # Inner Attrs Class

                # fastavro does not allow you to redefine the same record type more than once,
                # so only define an attr record once
                if attribute.type.__name__ in self.schema_record_names:
                    inner_field = {
                        "name": attribute.name,
                        "type": attribute.type.__name__,
                    }
                else:
                    self.schema_record_names.add(attribute.type.__name__)
                    inner_field = self._record_fields_for_attrs_class(
                        attribute.type, attribute.name
                    )
            # else attribute is an costom type and
            # there needs to be AvroAttrsBridgeExtension for attribute in self.extensions
            else:
                inner_field = None
                extension = self.extensions.get(attribute.type)
                if extension is not None:
                    inner_field = {
                        "name": attribute.name,
                        "type": extension.record_fields(),
                    }
                else:
                    raise TypeError(
                        f"AvroAttrsBridgeExtension for {attribute.type} not in self.extensions."
                    )
            # Assume attribute is optional if it has a default value
            # The default value is always set to None to allow attr class to handle dealing with default values
            # in dict_to_attrs function in this class
            if attribute.default is not attr.NOTHING:
                inner_field["type"] = ["null", inner_field["type"]]
                inner_field["default"] = None
            field["type"]["fields"].append(inner_field)

        return field

    def to_dict(self, obj, event_overrides=None):
        """
        Convert obj into dictionary that matches avro schema (self.schema).

        Args:
            obj: instance of self._attr_cls
            event_overrides: dict with following value overwrites:
                - id: unique id for this event. If id is not in dict, a uuid1 will be created for this event
                      For more info, see OEP-41: Asynchronous Server Event Message Format
                - time: time stamp for this event. If time is not in dict, datetime.now() will be called
                        For more info, see OEP-41: Asynchronous Server Event Message Format
        """
        if isinstance(event_overrides, dict) and "id" in event_overrides:
            event_id = event_overrides["id"]
        else:
            event_id = str(uuid.uuid1())
        if isinstance(event_overrides, dict) and "time" in event_overrides:
            event_timestamp = event_overrides["time"]
        else:
            event_timestamp = datetime.now().isoformat()
        obj_as_dict = attr.asdict(obj, value_serializer=self._extension_serializer)
        # Not sure if it makes sense to keep version info here since the schema registry will actually
        # keep track of versions and the topic can have only one associated schema at a time.
        avro_record = dict(
            id=event_id,
            type=self.config["type"],
            time=event_timestamp,
            source=self.config["source"],
            sourcehost=self.config["sourcehost"],
            minorversion=0,
            data=obj_as_dict,
        )
        return avro_record

    def serialize(self, obj) -> bytes:
        """
        Convert from attrs to a valid avro record.
        """
        avro_record = self.to_dict(obj)
        # Try to serialize using the generated schema.
        out = io.BytesIO()
        fastavro.schemaless_writer(out, self.schema_dict, avro_record)
        out.seek(0)
        return out.read()

    def _extension_serializer(self, _, field, value):
        """
        Pass this callback into attrs.asdict function as "value_serializer" arg.

        Serializes values for which an extention exists in self.extensions dict.
        """
        extension = self.extensions.get(field.type, None)
        if extension is not None:
            return extension.serialize(value)
        return value

    def deserialize(self, data: bytes, writer_schema=None) -> object:
        """
        Deserialize data into self.attrs_cls instance.

        Args:
            data: bytes that you want to deserialize
            writer_schema: pass the schema used to serialize data if it is differnt from current schema
        """
        data_file = io.BytesIO(data)
        if writer_schema is not None:
            record = fastavro.schemaless_reader(
                data_file, writer_schema, self.schema_dict
            )
        else:
            record = fastavro.schemaless_reader(data_file, self.schema_dict)
        return self.dict_to_attrs(record["data"], self._attrs_cls)

    def dict_to_attrs(self, data: dict, attrs_cls):
        """
        Convert data into instantiated object of attrs_cls.
        """
        for attribute in attrs_cls.__attrs_attrs__:
            if attribute.name in data:
                sub_data = data[attribute.name]
                if sub_data is None:
                    # delete keys that have defaults in attr class and which have None as value
                    # this is to let attr class take care of creating default values
                    if attribute.default is not attr.NOTHING:
                        del data[attribute.name]
                else:
                    if hasattr(attribute.type, "__attrs_attrs__"):
                        if attribute.name in data:
                            data[attribute.name] = self.dict_to_attrs(
                                sub_data, attribute.type
                            )
                    elif attribute.type in self.extensions:
                        extension = self.extensions.get(attribute.type)
                        if attribute.name in data:
                            data[attribute.name] = extension.deserialize(sub_data)
                        else:
                            raise Exception(
                                f"Necessary key: {attribute.name} not found in data dict"
                            )
                    elif attribute.type not in AVRO_TYPE_FOR:
                        raise TypeError(
                            f"Unable to deserialize {attribute.type} data, please add extension for custom data type"
                        )

        return attrs_cls(**data)


class AvroAttrsBridgeKafkaWrapper(AvroAttrsBridge):
    """
    Wrapper class to help AvroAttrsBridge to work with kafka.
    """

    def to_dict(self, obj, _kafka_context):  # pylint: disable=signature-differs
        """
        Pass this function as callable input to confluent_kafka::AvroSerializer.
        """
        return super().to_dict(obj)

    def from_dict(self, data, _kafka_context):
        """
        Pass this function as callable input to confluent_kafka::AvroDeSerializer.
        """
        return self.dict_to_attrs(data["data"], self._attrs_cls)



"""
Tests for AvroAttrsBridge.
"""
from datetime import datetime
from unittest import TestCase

import attr
from opaque_keys.edx.keys import CourseKey

from openedx_events.bridge.avro_attrs_bridge import AvroAttrsBridge
from openedx_events.bridge.avro_attrs_bridge_extensions import CourseKeyAvroAttrsBridgeExtension
from openedx_events.learning.data import CourseData, CourseEnrollmentData, UserData, UserPersonalData


class TestNoneBaseTypesInBridge(TestCase):
    """
    Tests to make sure AttrsAvroBridge handles custom types correctly.
    """

    def setUp(self):
        super().setUp()
        user_personal_data = UserPersonalData(
            username="username", email="email", name="name"
        )
        user_data = UserData(id=1, is_active=True, pii=user_personal_data)
        # define Coursedata, which needs Coursekey, which needs opaque key
        course_id = "course-v1:edX+DemoX.1+2014"
        course_key = CourseKey.from_string(course_id)
        course_data = CourseData(
            course_key=course_key,
            display_name="display_name",
            start=datetime.now(),
            end=datetime.now(),
        )
        self.course_enrollment_data = CourseEnrollmentData(
            user=user_data,
            course=course_data,
            mode="mode",
            is_active=False,
            creation_date=datetime.now(),
            created_by=user_data,
        )

    def test_non_base_types_in_bridge(self):
        """
        Test to makes ure AvroattrsBridge works correctly with non-attr classes.

        Specifically, testing to make sure the extension classes work as intended.
        """
        bridge = AvroAttrsBridge(
            CourseEnrollmentData,
            extensions={
                CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
            },
        )
        serialized_course_enrollment_data = bridge.serialize(
            self.course_enrollment_data
        )

        object_from_wire = bridge.deserialize(serialized_course_enrollment_data)
        assert self.course_enrollment_data == object_from_wire

    def test_schema_evolution_add_value(self):
        # Create object from old specification
        old_bridge = AvroAttrsBridge(
            CourseEnrollmentData,
            extensions={
                CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
            },
        )
        serialized_course_enrollment_data = old_bridge.serialize(
            self.course_enrollment_data
        )

        def inner_scope(self):
            @attr.s(frozen=True)
            class CourseEnrollmentData:  # pylint: disable=redefined-outer-name
                """
                Temp class create to test schema evolution.
                """

                user = attr.ib(type=UserData)
                course = attr.ib(type=CourseData)
                mode = attr.ib(type=str)
                is_active = attr.ib(type=bool)
                creation_date = attr.ib(type=datetime)
                is_active_2 = attr.ib(type=bool, default=False)
                created_by = attr.ib(type=UserData, default=None)

            new_bridge = AvroAttrsBridge(
                CourseEnrollmentData,
                extensions={
                    CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
                },
            )
            object_from_wire_as_dict = attr.asdict(
                new_bridge.deserialize(
                    serialized_course_enrollment_data, old_bridge.schema_dict
                )
            )

            original_object_as_dict = attr.asdict(self.course_enrollment_data)
            original_object_as_dict["is_active_2"] = False
            assert object_from_wire_as_dict == original_object_as_dict

        inner_scope(self)

    def test_schema_evolution_remove_value(self):
        # Create object from old specification
        old_bridge = AvroAttrsBridge(
            CourseEnrollmentData,
            extensions={
                CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
            },
        )
        serialized_course_enrollment_data = old_bridge.serialize(
            self.course_enrollment_data
        )

        def inner_scope(self):
            @attr.s(frozen=True)
            class CourseEnrollmentData:  # pylint: disable=redefined-outer-name
                """
                Temp class create to test schema evolution.
                """

                user = attr.ib(type=UserData)
                course = attr.ib(type=CourseData)
                mode = attr.ib(type=str)
                creation_date = attr.ib(type=datetime)
                created_by = attr.ib(type=UserData, default=None)

            new_bridge = AvroAttrsBridge(
                CourseEnrollmentData,
                extensions={
                    CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
                },
            )
            object_from_wire_as_dict = attr.asdict(
                new_bridge.deserialize(
                    serialized_course_enrollment_data, old_bridge.schema_dict
                )
            )

            original_object_as_dict = attr.asdict(self.course_enrollment_data)
            del original_object_as_dict["is_active"]
            assert object_from_wire_as_dict == original_object_as_dict

        inner_scope(self)

    def test_schema_evolution_add_complex_value(self):
        # Create object from old specification
        old_bridge = AvroAttrsBridge(
            CourseEnrollmentData,
            extensions={
                CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
            },
        )
        serialized_course_enrollment_data = old_bridge.serialize(
            self.course_enrollment_data
        )

        def inner_scope(self):

            user_personal_data = UserPersonalData(
                username="username", email="email", name="name"
            )
            user_data = UserData(id=1, is_active=True, pii=user_personal_data)

            @attr.s(frozen=True)
            class CourseEnrollmentData:  # pylint: disable=redefined-outer-name
                """
                Temp class create to test schema evolution.
                """

                user = attr.ib(type=UserData)
                course = attr.ib(type=CourseData)
                mode = attr.ib(type=str)
                is_active = attr.ib(type=bool)
                creation_date = attr.ib(type=datetime)
                created_by = attr.ib(type=UserData, default=None)
                user2 = attr.ib(type=UserData, default=user_data)

            new_bridge = AvroAttrsBridge(
                CourseEnrollmentData,
                extensions={
                    CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
                },
            )
            object_from_wire_as_dict = attr.asdict(
                new_bridge.deserialize(
                    serialized_course_enrollment_data, old_bridge.schema_dict
                )
            )

            original_object_as_dict = attr.asdict(self.course_enrollment_data)
            original_object_as_dict["user2"] = attr.asdict(user_data)
            assert object_from_wire_as_dict == original_object_as_dict

        inner_scope(self)

    def test_schema_evolution_add_complex_extension_value(self):
        # Create object from old specification
        old_bridge = AvroAttrsBridge(
            CourseEnrollmentData,
            extensions={
                CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
            },
        )
        serialized_course_enrollment_data = old_bridge.serialize(
            self.course_enrollment_data
        )

        def inner_scope(self):

            @attr.s(frozen=True)
            class CourseEnrollmentData:  # pylint: disable=redefined-outer-name
                """
                Temp class create to test schema evolution.
                """

                user = attr.ib(type=UserData)
                course = attr.ib(type=CourseData)
                mode = attr.ib(type=str)
                is_active = attr.ib(type=bool)
                creation_date = attr.ib(type=datetime)
                created_by = attr.ib(type=UserData, default=None)
                added_date = attr.ib(type=datetime, default=None)

            new_bridge = AvroAttrsBridge(
                CourseEnrollmentData,
                extensions={
                    CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
                },
            )
            object_from_wire_as_dict = attr.asdict(
                new_bridge.deserialize(
                    serialized_course_enrollment_data, old_bridge.schema_dict
                )
            )
            original_object_as_dict = attr.asdict(self.course_enrollment_data)
            original_object_as_dict["added_date"] = None
            assert object_from_wire_as_dict == original_object_as_dict

        inner_scope(self)

    def test_throw_exception_on_unextended_custom_type(self):
        with self.assertRaises(TypeError):
            # CourseEnrollmentData has CourseKey and datetime as custom types
            # This should raise TypeError cause no extensions are being passed to bridge
            AvroAttrsBridge(CourseEnrollmentData)


def test_object_evolution_add_value():
    @attr.s(auto_attribs=True)
    class TestData:
        sub_name: str
        course_id: str

    original_bridge = AvroAttrsBridge(TestData)

    record = TestData("sub_name", "course_id")
    serialized_record = original_bridge.serialize(record)

    @attr.s(auto_attribs=True)
    class TestData:  # pylint: disable=function-redefined
        sub_name: str
        course_id: str
        added_key: str = "default_value"

    new_bridge = AvroAttrsBridge(TestData)
    deserialized_obj = new_bridge.deserialize(
        serialized_record, original_bridge.schema_dict
    )
    assert deserialized_obj == TestData(
        sub_name="sub_name", course_id="course_id", added_key="default_value"
    )


def test_object_evolution_remove_value():
    @attr.s(auto_attribs=True)
    class TestData:
        sub_name: str
        course_id: str
        removed_key: str

    original_bridge = AvroAttrsBridge(TestData)

    record = TestData("sub_name", "course_id", "removed_value")
    serialized_record = original_bridge.serialize(record)

    @attr.s(auto_attribs=True)
    class TestData:  # pylint: disable=function-redefined
        sub_name: str
        course_id: str

    new_bridge = AvroAttrsBridge(TestData)
    deserialized_obj = new_bridge.deserialize(
        serialized_record, original_bridge.schema_dict
    )
    assert deserialized_obj == TestData(sub_name="sub_name", course_id="course_id")


def test_base_types():
    @attr.s(auto_attribs=True)
    class SubTestData:
        sub_name: str
        course_id: str

    @attr.s(auto_attribs=True)
    class SubTestData2:
        sub_name: str
        course_id: str

    @attr.s(auto_attribs=True)
    class TestData:  # pylint: disable=missing-class-docstring
        name: str
        course_id: str
        user_id: int
        sub_test: SubTestData
        uber_sub_test: SubTestData2

    bridge = AvroAttrsBridge(TestData)

    # A test record that we can try to serialize to avro.
    record = TestData(
        "foo",
        "bar.course",
        1,
        SubTestData("a.sub.name", "a.nother.course"),
        SubTestData2("b.uber.sub.name", "b.uber.another.course"),
    )
    serialized_record = bridge.serialize(record)

    # Try to de-serialize back to an attrs class.
    object_from_wire = bridge.deserialize(serialized_record)
    assert record == object_from_wire
